/* bqd.c - Binary encoding for basic Q data types (C source)
   Copyright (C) 2005  Peter Minten

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "bqd.h"

// debug helper
void p(char* str)
{
  printf("%s\n",str);
}

MODULE(bqd)

//from clib.c - needed for bytestr manipulation

typedef struct bstr {
  long size;
  unsigned char *v;
} bstr_t;

static expr mkbstr(long size, void *v)
{
  bstr_t *m;
  if ((m = malloc(sizeof(bstr_t)))) {
    m->size = size;
    m->v = (unsigned char*)v;
    return mkobj(type(ByteStr), m);
  } else {
    if (v) free(v);
    return __ERROR;
  }
}

//end from clib.c

//The two exposed functions

FUNCTION(bqd,bqd_encode,argc,argv)
{
  char *v;
  long size;
  long used;

  if(argc == 1)
  {
    v = (char*) malloc(BQD_MEM_INITIAL_SIZE);
    size = BQD_MEM_INITIAL_SIZE;
    v[0] = BQD_VERSION;
    used = 1;
    
    if(!bytify(argv[0], &v, &size, &used))
    {
      return mkbstr(used, v);
    }
    else
    {
      return __ERROR;
    }
  }
  else
  {
    return __ERROR; //too few or too much arguments
  }
}

FUNCTION(bqd,bqd_decode,argc,argv)
{
  bstr_t *bstr;
  long pos;

  if(argc == 1 && isobj(argv[0], type(ByteStr), (void**)&bstr))
  {
    if(bstr->v[0] != BQD_VERSION)
      return __ERROR;
    else
    {
      pos = 1;
      return unbytify(bstr->v, &pos);
    }
  }
  else
  {
    return __ERROR; //too few or too much arguments
  }  
}

// internals
int bytify(expr e, char **v, long *size, long *used)
{
  long vint;
  unsigned long vuint;
  mpz_t vmpz;
  double vfloat;
  char * vstr;
  expr * vtuple;
  expr head, tail, tail2;
  unsigned int count, countn;
  
  // First, make sure there is room for at least an int
  if(require_mem(v,size,*used,sizeof(int)+1) == -1)
    return -1;
  
  if(istrue(e))
  {
    (*v)[*used] = CODE_TRUE;
    (*used)++;
  }
  else if(isfalse(e))
  {
    (*v)[*used] = CODE_FALSE;
    (*used)++;
  }
  else if(isnil(e))
  {
    (*v)[*used] = CODE_FALSE;
    (*used)++;
  }
  else if(isint(e,&vint))
  {
    // NOTE: not sure if this works well when encoding on 64 bits machine
    // and decoding on 32 bits machine
    unsigned int temp;
    (*v)[*used] = CODE_INT;
    (*used)++;
    temp = htonl(vint);
    memcpy(*v+*used,&temp,sizeof(int));
    (*used) += sizeof(int);
  }
  else if(isuint(e,&vuint))
  {
    unsigned int temp;
    (*v)[*used] = CODE_UINT;
    (*used)++;
    temp = htonl(vuint);
    memcpy(*v+*used,&temp,sizeof(unsigned int));
    (*used) += sizeof(unsigned int);
  }
  else if(ismpz(e,&vmpz))
  {
    FILE * stream;
    unsigned int length, lengthn;
    (*v)[*used] = CODE_MPZ;
    (*used)++;
    // Ok, this is a little tricky, basically we first overallocate the
    // buffer and write the mpz value into it.
    if(require_mem(v,size,*used,BQD_STRING_STREAM_SIZE) == -1) return -1;
    stream = (FILE*) fmemopen(*v+*used+sizeof(int),BQD_STRING_STREAM_SIZE,"w");
    if(!stream) return -1;
    if(!(length = mpz_out_raw(stream,vmpz))) return -1;
    lengthn = htonl(length); //prevent endianness problems with lengths
    memcpy(*v+*used,&lengthn,sizeof(int)); //write length
    (*used) += sizeof(int) + length;
    fclose(stream);
  }
  else if(isfloat(e,&vfloat))
  {
    if(require_mem(v,size,*used,sizeof(float)+1) == -1) return -1;
    (*v)[*used] = CODE_FLOAT;
    (*used)++;
    memcpy(*v+*used,&vfloat,sizeof(double));
    (*used) += sizeof(double);
  }
  else if(isstr(e,&vstr))
  {
    unsigned int length, lengthn;
    (*v)[*used] = CODE_STRING;
    (*used)++;
    length = strlen(vstr);
    if(require_mem(v,size,*used,length) == -1) return -1;
    lengthn = htonl(length);
    memcpy(*v+*used,&lengthn,sizeof(int));
    (*used) += sizeof(int);
    memcpy(*v+*used,vstr,length);
    (*used) += length;
  }
  else if(iscons(e,&head,&tail))
  {
    (*v)[*used] = CODE_LIST;
    (*used)++;
    //NOTE: as we don't know the count before we've traversed the list it will
    //have to be written at the end
    unsigned int countloc = *used;
    (*used) += sizeof(int); //reserve that spot

    bytify(head,v,size,used);
    count = 1;

    tail2 = tail; //needed because otherwise we'd get islist(tail,&head,&tail)
    while(iscons(tail2,&head,&tail)) //very evil, hope it works
    {
      bytify(head,v,size,used);
      tail2 = tail;
      count++;
    }

    countn = htonl(count);
    memcpy(*v + countloc,&countn,sizeof(int));
  }
  else if(istuple(e,&count,&vtuple))
  {
    int i;
    (*v)[*used] = CODE_TUPLE;
    (*used)++;
    countn = htonl(count);
    memcpy(*v+*used,&countn,sizeof(int));
    (*used) += sizeof(int);
    for(i = 0; i < count; i++)
      bytify(vtuple[i],v,size,used);
  }
  else
  {
    return -1; //invalid type
  }
  
  return 0;
}

expr unbytify(char *v, long *pos)
{
  int vint;
  unsigned int vuint;
  mpz_t vmpz;
  double vfloat;
  char * vstr;
  unsigned int length, lengthn, count, countn;
  FILE * stream;
  int i;
  expr *varr;

  switch(v[*pos])
  {
  case CODE_TRUE:
    (*pos)++;
    return mktrue;
  case CODE_FALSE:
    (*pos)++;
    return mkfalse;
  case CODE_NIL:
    (*pos)++;
    return mknil;
  case CODE_INT:
    (*pos)++;
    memcpy(&vint,v+(*pos),sizeof(int));
    (*pos) += sizeof(int);
    return mkint(ntohl(vint));
  case CODE_UINT:
    (*pos)++;
    memcpy(&vuint,v+(*pos),sizeof(unsigned int));
    (*pos) += sizeof(unsigned int);
    return mkuint(ntohl(vuint));
  case CODE_MPZ:
    (*pos)++;
    memcpy(&lengthn,v+(*pos),sizeof(int));   
    length = ntohl(lengthn);
    (*pos) += sizeof(int);
    stream = (FILE*) fmemopen(v+(*pos),length,"r");
    if(!stream) return __ERROR;
    mpz_init(vmpz);
    if(!mpz_inp_raw(vmpz, stream)) return __ERROR;
    fclose(stream);
    (*pos)+=length;
    return mkmpz((void*) vmpz);
  case CODE_FLOAT:
    (*pos)++;
    memcpy(&vfloat,v+(*pos),sizeof(double));
    (*pos) += sizeof(double);
    return mkfloat(vfloat);
  case CODE_STRING:
    (*pos)++;
    memcpy(&lengthn,v+(*pos),sizeof(int));   
    length = ntohl(lengthn);
    (*pos) += sizeof(int);
    vstr = (char*) malloc(length+1);
    memcpy(vstr, v+(*pos), length);
    vstr[length] = '\0'; //add terminator
    (*pos) += length;
    return mkstr(vstr); //note that we can't free vstr as it's taken over
  case CODE_LIST:
    (*pos)++;
    memcpy(&countn,v+(*pos),sizeof(int));   
    count = ntohl(countn);
    (*pos) += sizeof(int);
    varr = malloc(count * sizeof(expr));
    for(i = 0; i < count; i++)
    {
      varr[i] = unbytify(v,pos);
    }
    return mklistv(count,varr);
  case CODE_TUPLE:
    (*pos)++;
    memcpy(&countn,v+(*pos),sizeof(int));   
    count = ntohl(countn);
    (*pos) += sizeof(int);
    varr = malloc(count * sizeof(expr));
    for(i = 0; i < count; i++)
    {
      varr[i] = unbytify(v,pos);
    }
    return mktuplev(count,varr);
  default:
    return __ERROR;
  }
}

// helpers

// Allocate extra memory if needed. Inlined to avoid the overhead
// of a function call, reasonable as it's a rather small function.
inline int require_mem(char **v, long *size, long used, long needed)
{
  if(*size < used + needed)
  {
    *size = *size + (BQD_MEM_STEP_SIZE > needed ? BQD_MEM_STEP_SIZE :
                     ((needed / BQD_MEM_STEP_SIZE) + 1) * BQD_MEM_STEP_SIZE);
    *v = realloc(*v, *size);
    //    printf("UPDATE: size = %d, used = %d\n", *size, used);
    return (*v == NULL ? -1 : 0); 
  }
  else
  {
    return 0;
  }
}
