/* bqd.q - Binary encoding for basic Q data types (old Q version)
   Copyright (C) 2005  Peter Minten

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//BQD = Binary Q Data, a simple binary encoding for basic Q data types.
//
// Each BQD (byte) string starts with one byte which denotes the BQD 
// version.
//
// A BQD string consists of one object (which could contain others).
// The type of object is denoted by a type tag (a byte). The following
// codes are implemented:
// 0 = (), 1 = true, 2 = false, 3 = Int, 4 = BigInt, 5 = Float, 
// 6 = String, 7 = List, 8 = Tuple
//
// Lengths (for BigInt, String, List and Tuple) are denoted with special 
// length tags. These are 1 byte long for lengths < 64 and 5 bytes long
// for other lengths. Note that the most significant bit of the first byte
// is used as the extension bit, if it's on the number is four bytes long
// (starting after the first byte). The next most signicant bit of the first
// byte is used as sign for big int's.

def BQD_VERSION = 1;

public type BqdException : Exception = special const bqd_exception X;

// The external interfaces.
public bqd_encode2 Object, bqd_decode2 ByteStr;

bqd_encode2 O = bcat [bytestr [BQD_VERSION], bqdify O];

bqd_decode2 B:ByteStr = (unbqdify T B 2)!0
                      if V = BQD_VERSION
                      where 
                        V = byte 0 B,
                        T = byte 1 B;
bqd_decode2 B:ByteStr = throw (bqd_exception "Wrong BQD version number.") otherwise;

// Helpers

// bytify_length Length Sign (true if negative)
bytify_length L S = bytestr [L or (ifelse S 0x40 0)] if L < 64;
                  = bcat [bytestr [0x80 or (ifelse S 0x40 0)],
                          bytestr L] if L < 0xFFFFFFFF;
                  = throw (bqd_exception "Too long length.") otherwise;

// Returns (Length, Sign, NewPos)
unbytify_length B P = ((B!P and 0x3F),(ifelse ((B!P and 0x40) = 0x40) true false),
                       P+1) if ((B!P) and 0x80) = 0;
                    = (bint (bsub B (P+1) (P+4)),
                       (ifelse ((B!P and 0x40) = 0x40) true false), P+5) otherwise;

// BQDification
bqdify () = bytestr [0];
bqdify true = bytestr [1];
bqdify false = bytestr [2];

bqdify I:Int = bcat [bytestr [3], bytestr I] if (I <= 0xFFFFFFFF) and (I >= 0);
             = bcat [bytestr [4], bytify_length (bsize B) (I < 0), B]
                 if (bsize B) <= 127
                 where
                   B = bytestr (abs I);
             = throw (bqd_exception "Too large or small integer.") otherwise;

bqdify F:Float = bcat [bytestr [5], bytestr F];

bqdify S:String = bcat [bytestr [6], bytify_length (bsize B) false, B]
                    where
                      B = bytestr S;

bqdify L:List = bcat [bytestr [7], bytify_length (#L) false, B]
                  where
                    B = bqdify_list L [];

// Note that lists are actually stored reversed, saves O(list_size*2) 
// in reverse operations.
bqdify_list [] A =  bcat A;
bqdify_list [H|T] A = bqdify_list T [bqdify H|A];

bqdify T:Tuple = bcat [bytestr [8], bytify_length (#T) false, B]
                   where
                     B = bqdify_list (list T) [];

// UnBQDification

// Returns (Object, NewPos).
unbqdify 0 B P = ((), P);
unbqdify 1 B P = (true, P);
unbqdify 2 B P = (false, P);

unbqdify 3 B P = ((bint (bsub B P (P+3))), P+4);
unbqdify 4 B P = (F * (bint (bsub B NP (NP+L-1))), NP+L)
                   where 
                     (L,S,NP) = (unbytify_length B P),
                     F = ifelse S (-1) 1;

unbqdify 5 B P = ((bfloat (bsub B P (P+7))), P+8);
unbqdify 6 B P = (bstr (bsub B NP (NP+L-1)), NP+L)
                   where 
                     (L,_,NP) = unbytify_length B P;

unbqdify 7 B P = (NA,NNP)
                   where
                     (L, _, NP) = unbytify_length B P,
                     (NA, NNP) = unbqdify_list B NP L [];

unbqdify_list B P C A = (A,P) if C = 0;
                      = unbqdify_list B NP (C-1) [NA|A]
                          where
                            (NA, NP) = unbqdify (B!P) B (P+1)
                          otherwise;

unbqdify 8 B P = (tuple NA, NP)
                   where
                     (NA, NP) = unbqdify 7 B P;   // tuples are stored like lists